import torch
from torch import Tensor
from torch.nn import Module


class SinPositionEmbedding(Module):
    def __init__(
        self, num_embeddings: int, embedding_dim: int, sin_scale_factor: int = 30
    ) -> None:
        self._check_embedding_dim_is_even(embedding_dim)
        super().__init__()

        position_embedding_matrix = self._compute_position_embedding_matrix(
            num_embeddings, embedding_dim, sin_scale_factor
        )
        self.register_buffer("position_embedding_matrix", position_embedding_matrix)

    def forward(self, token_indices: Tensor) -> Tensor:
        return self.position_embedding_matrix[token_indices]

    def _check_embedding_dim_is_even(self, embedding_dim: int) -> None:
        if embedding_dim % 2 != 0:
            raise RuntimeError("embedding_dim must be even")

    def _compute_position_embedding_matrix(
        self, num_embeddings: int, embedding_dim: int, sin_scale_factor: int
    ) -> Tensor:
        num_embeddings_to_represent_including_padding = num_embeddings + 1
        position_embedding_matrix = torch.zeros(
            num_embeddings_to_represent_including_padding, embedding_dim
        )

        position_index_column_vector = torch.arange(0, num_embeddings).unsqueeze(1)
        wavelength_scaling_row_vector = self._compute_wavelength_scaling_row_vector(
            embedding_dim, sin_scale_factor
        )

        position_embedding_matrix[1:, 0::2] = torch.sin(
            position_index_column_vector / wavelength_scaling_row_vector
        )
        position_embedding_matrix[1:, 1::2] = torch.cos(
            position_index_column_vector / wavelength_scaling_row_vector
        )

        return position_embedding_matrix

    def _compute_wavelength_scaling_row_vector(
        self, embedding_dim: int, sin_scale_factor: int
    ) -> Tensor:
        scaling_strength_for_every_pair_of_embedding_dims = (
            torch.arange(0, embedding_dim, 2) / embedding_dim
        )
        scaling_factor_for_every_pair_of_embedding_dims = torch.pow(
            sin_scale_factor, scaling_strength_for_every_pair_of_embedding_dims
        )

        return scaling_factor_for_every_pair_of_embedding_dims



import torch
from torch import testing
from libtcrlm.token_embedder.sin_position_embedding import SinPositionEmbedding


def test_embedding():
    embedding = SinPositionEmbedding(num_embeddings=10, embedding_dim=4)
    mock_batch_of_token_indices = torch.tensor([[1, 2], [1, 0]], dtype=torch.long)

    result = embedding.forward(mock_batch_of_token_indices)
    expected = torch.tensor(
        [
            [[0.0000, 1.0000, 0.0000, 1.0000], [0.8415, 0.5403, 0.1816, 0.9834]],
            [[0.0000, 1.0000, 0.0000, 1.0000], [0.0000, 0.0000, 0.0000, 0.0000]],
        ]
    )

    testing.assert_close(result, expected, rtol=0, atol=0.00005)



from abc import ABC, abstractmethod
from torch import Tensor
from torch.nn import Linear, Module, TransformerEncoder, TransformerEncoderLayer
from typing import Optional


class SelfAttentionStack(ABC, Module):
    @property
    @abstractmethod
    def d_model(self) -> int:
        pass

    @abstractmethod
    def forward(self, token_embeddings: Tensor, padding_mask: Tensor) -> Tensor:
        pass

    @abstractmethod
    def get_token_embeddings_at_penultimate_layer(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        pass

    @abstractmethod
    def set_fine_tuning_mode(self, turn_on: bool) -> None:
        pass


class SelfAttentionStackWithBuiltins(SelfAttentionStack):
    d_model: int = None

    def __init__(
        self,
        num_layers: int,
        d_model: int,
        nhead: int,
        dim_feedforward: Optional[int] = None,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()

        if dim_feedforward is None:
            dim_feedforward = d_model * 4  # backwards compatibility

        self.d_model = d_model
        self._num_layers_in_stack = num_layers

        self_attention_block = TransformerEncoderLayer(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            batch_first=True,
        )
        self._self_attention_stack = TransformerEncoder(
            encoder_layer=self_attention_block, num_layers=num_layers
        )

    def forward(self, token_embeddings: Tensor, padding_mask: Tensor) -> Tensor:
        return self._self_attention_stack.forward(
            src=token_embeddings, src_key_padding_mask=padding_mask
        )

    def get_token_embeddings_at_penultimate_layer(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        penultimate_layer_index = self._num_layers_in_stack - 1

        for layer in self._self_attention_stack.layers[:penultimate_layer_index]:
            token_embeddings = layer.forward(
                src=token_embeddings, src_key_padding_mask=padding_mask
            )

        return token_embeddings

    def set_fine_tuning_mode(self, turn_on: bool) -> None:
        upper_layers_require_grad = not turn_on
        penultimate_layer_index = self._num_layers_in_stack - 1

        for layer in self._self_attention_stack.layers[:penultimate_layer_index]:
            layer.requires_grad_(upper_layers_require_grad)


class SelfAttentionStackWithInitialProjection(SelfAttentionStack):
    d_model: int = None

    def __init__(
        self,
        num_layers: int,
        embedding_dim: int,
        d_model: int,
        nhead: int,
        dim_feedforward: Optional[int] = None,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()

        self.d_model = d_model

        self._initial_projector = Linear(
            in_features=embedding_dim, out_features=d_model, bias=False
        )
        self._standard_stack = SelfAttentionStackWithBuiltins(
            num_layers=num_layers,
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
        )

    def forward(self, token_embeddings: Tensor, padding_mask: Tensor) -> Tensor:
        projected_embeddings = self._initial_projector.forward(token_embeddings)
        return self._standard_stack.forward(projected_embeddings, padding_mask)

    def get_token_embeddings_at_penultimate_layer(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        projected_embeddings = self._initial_projector.forward(token_embeddings)
        return self._standard_stack.get_token_embeddings_at_penultimate_layer(
            projected_embeddings, padding_mask
        )

    def set_fine_tuning_mode(self, turn_on: bool) -> None:
        self._standard_stack.set_fine_tuning_mode(turn_on)



from abc import ABC, abstractmethod
from enum import IntEnum
from libtcrlm.schema import Tcr
from torch import Tensor


class Tokeniser(ABC):
    @property
    @abstractmethod
    def token_vocabulary_index(self) -> IntEnum:
        pass

    @abstractmethod
    def tokenise(self, tcr: Tcr) -> Tensor:
        pass



from libtcrlm.tokeniser.tokeniser import Tokeniser
from libtcrlm.tokeniser.token_indices import (
    AminoAcidTokenIndex,
    CdrCompartmentIndex,
    SingleChainCdrCompartmentIndex,
)
from libtcrlm.schema import Tcr
import torch
from torch import Tensor
from typing import List, Optional, Tuple


class CdrTokeniser(Tokeniser):
    """
    Tokenise TCR in terms of its alpha and beta chain CDRs 1 2 and 3.

    Dim 0: token index
    Dim 1: topen position
    Dim 2: CDR length
    Dim 3: CDR index
    """

    token_vocabulary_index = AminoAcidTokenIndex

    def tokenise(self, tcr: Tcr) -> Tensor:
        initial_cls_vector = (AminoAcidTokenIndex.CLS, 0, 0, CdrCompartmentIndex.NULL)

        cdr1a = self._convert_to_numerical_form(
            tcr.cdr1a_sequence, CdrCompartmentIndex.CDR1A
        )
        cdr2a = self._convert_to_numerical_form(
            tcr.cdr2a_sequence, CdrCompartmentIndex.CDR2A
        )
        cdr3a = self._convert_to_numerical_form(
            tcr.junction_a_sequence, CdrCompartmentIndex.CDR3A
        )

        cdr1b = self._convert_to_numerical_form(
            tcr.cdr1b_sequence, CdrCompartmentIndex.CDR1B
        )
        cdr2b = self._convert_to_numerical_form(
            tcr.cdr2b_sequence, CdrCompartmentIndex.CDR2B
        )
        cdr3b = self._convert_to_numerical_form(
            tcr.junction_b_sequence, CdrCompartmentIndex.CDR3B
        )

        all_cdrs_tokenised = (
            [initial_cls_vector] + cdr1a + cdr2a + cdr3a + cdr1b + cdr2b + cdr3b
        )

        number_of_tokens_other_than_initial_cls = len(all_cdrs_tokenised) - 1
        if number_of_tokens_other_than_initial_cls == 0:
            raise RuntimeError(f"tcr {tcr} does not contain any TCR information")

        return torch.tensor(all_cdrs_tokenised, dtype=torch.long)

    def _convert_to_numerical_form(
        self, aa_sequence: Optional[str], cdr_index: CdrCompartmentIndex
    ) -> List[Tuple[int]]:
        if aa_sequence is None:
            return []

        token_indices = [AminoAcidTokenIndex[aa] for aa in aa_sequence]
        token_positions = [idx for idx, _ in enumerate(aa_sequence, start=1)]
        cdr_length = [len(aa_sequence) for _ in aa_sequence]
        compartment_index = [cdr_index for _ in aa_sequence]

        iterator_over_token_vectors = zip(
            token_indices, token_positions, cdr_length, compartment_index
        )

        return list(iterator_over_token_vectors)


class AlphaCdrTokeniser(Tokeniser):
    """
    Tokenise TCR in terms of its alpha chain CDRs 1 2 and 3.

    Dim 0: token index
    Dim 1: topen position
    Dim 2: CDR length
    Dim 3: CDR index
    """

    token_vocabulary_index = AminoAcidTokenIndex

    def tokenise(self, tcr: Tcr) -> Tensor:
        initial_cls_vector = (
            AminoAcidTokenIndex.CLS,
            0,
            0,
            SingleChainCdrCompartmentIndex.NULL,
        )

        cdr1b = self._convert_to_numerical_form(
            tcr.cdr1a_sequence, SingleChainCdrCompartmentIndex.CDR1
        )
        cdr2b = self._convert_to_numerical_form(
            tcr.cdr2a_sequence, SingleChainCdrCompartmentIndex.CDR2
        )
        cdr3b = self._convert_to_numerical_form(
            tcr.junction_a_sequence, SingleChainCdrCompartmentIndex.CDR3
        )

        all_cdrs_tokenised = [initial_cls_vector] + cdr1b + cdr2b + cdr3b

        number_of_tokens_other_than_initial_cls = len(all_cdrs_tokenised) - 1
        if number_of_tokens_other_than_initial_cls == 0:
            raise RuntimeError(f"tcr {tcr} does not contain any TRA information")

        return torch.tensor(all_cdrs_tokenised, dtype=torch.long)

    def _convert_to_numerical_form(
        self, aa_sequence: Optional[str], cdr_index: SingleChainCdrCompartmentIndex
    ) -> List[Tuple[int]]:
        if aa_sequence is None:
            return []

        token_indices = [AminoAcidTokenIndex[aa] for aa in aa_sequence]
        token_positions = [idx for idx, _ in enumerate(aa_sequence, start=1)]
        cdr_length = [len(aa_sequence) for _ in aa_sequence]
        compartment_index = [cdr_index for _ in aa_sequence]

        iterator_over_token_vectors = zip(
            token_indices, token_positions, cdr_length, compartment_index
        )

        return list(iterator_over_token_vectors)


class BetaCdrTokeniser(Tokeniser):
    """
    Tokenise TCR in terms of its beta chain CDRs 1 2 and 3.

    Dim 0: token index
    Dim 1: topen position
    Dim 2: CDR length
    Dim 3: CDR index
    """

    token_vocabulary_index = AminoAcidTokenIndex

    def tokenise(self, tcr: Tcr) -> Tensor:
        initial_cls_vector = (
            AminoAcidTokenIndex.CLS,
            0,
            0,
            SingleChainCdrCompartmentIndex.NULL,
        )

        cdr1b = self._convert_to_numerical_form(
            tcr.cdr1b_sequence, SingleChainCdrCompartmentIndex.CDR1
        )
        cdr2b = self._convert_to_numerical_form(
            tcr.cdr2b_sequence, SingleChainCdrCompartmentIndex.CDR2
        )
        cdr3b = self._convert_to_numerical_form(
            tcr.junction_b_sequence, SingleChainCdrCompartmentIndex.CDR3
        )

        all_cdrs_tokenised = [initial_cls_vector] + cdr1b + cdr2b + cdr3b

        number_of_tokens_other_than_initial_cls = len(all_cdrs_tokenised) - 1
        if number_of_tokens_other_than_initial_cls == 0:
            raise RuntimeError(f"tcr {tcr} does not contain any TRB information")

        return torch.tensor(all_cdrs_tokenised, dtype=torch.long)

    def _convert_to_numerical_form(
        self, aa_sequence: Optional[str], cdr_index: SingleChainCdrCompartmentIndex
    ) -> List[Tuple[int]]:
        if aa_sequence is None:
            return []

        token_indices = [AminoAcidTokenIndex[aa] for aa in aa_sequence]
        token_positions = [idx for idx, _ in enumerate(aa_sequence, start=1)]
        cdr_length = [len(aa_sequence) for _ in aa_sequence]
        compartment_index = [cdr_index for _ in aa_sequence]

        iterator_over_token_vectors = zip(
            token_indices, token_positions, cdr_length, compartment_index
        )

        return list(iterator_over_token_vectors)



import copy
from enum import Enum
from libtcrlm.schema import exception
import logging
import re
import tidytcells as tt
from typing import Literal, Optional, Union


logger = logging.getLogger(__name__)


def get_v_gene_indices(gene_symbol):
    match = re.match(r"TR[AB]V([0-9]+)(-([0-9]+))?", gene_symbol)

    group_num = int(match.group(1))
    sub_num_if_any = 0 if match.group(3) is None else int(match.group(3))

    return (group_num, sub_num_if_any)


functional_travs = tt.tr.query(
    contains_pattern="TRAV", functionality="F", precision="gene"
)
functional_trbvs = tt.tr.query(
    contains_pattern="TRBV", functionality="F", precision="gene"
)


TravGene = Enum(
    "TravGene", sorted(functional_travs, key=get_v_gene_indices), module=__name__
)
TrbvGene = Enum(
    "TrbvGene", sorted(functional_trbvs, key=get_v_gene_indices), module=__name__
)


class Tcrv:
    def __init__(
        self, gene: Union[TravGene, TrbvGene, None], allele_num: Optional[int]
    ) -> None:
        self.gene = gene
        self.allele_num = allele_num

        if not self._gene_is_unknown() and self._allele_is_unknown():
            self._assume_first_allele()

    @property
    def cdr1_sequence(self) -> Optional[str]:
        if self._gene_is_unknown():
            return None

        allele_symbol = self.__repr__()
        cdr1 = tt.tr.get_aa_sequence(allele_symbol)["CDR1-IMGT"]
        return cdr1

    @property
    def cdr2_sequence(self) -> Optional[str]:
        if self._gene_is_unknown():
            return None

        allele_symbol = self.__repr__()
        aa_sequence_dictionary = tt.tr.get_aa_sequence(allele_symbol)

        if "CDR2-IMGT" not in aa_sequence_dictionary:
            return ""

        return aa_sequence_dictionary["CDR2-IMGT"]

    def __eq__(self, __value: object) -> bool:
        return self.gene == __value.gene and self.allele_num == __value.allele_num

    def __repr__(self) -> str:
        if self._gene_is_unknown():
            return "?"

        return f"{self.gene.name}*{self.allele_num:02d}"

    def _gene_is_unknown(self) -> bool:
        return self.gene is None

    def _allele_is_unknown(self) -> bool:
        return self.allele_num is None

    def _assume_first_allele(self) -> None:
        self.allele_num = 1


class Tcr:
    def __init__(
        self,
        trav: Tcrv,
        junction_a_sequence: Optional[str],
        trbv: Tcrv,
        junction_b_sequence: Optional[str],
    ) -> None:
        self._trav = trav
        self.junction_a_sequence = junction_a_sequence

        self._trbv = trbv
        self.junction_b_sequence = junction_b_sequence

    @property
    def cdr1a_sequence(self) -> Optional[str]:
        return self._trav.cdr1_sequence

    @property
    def cdr2a_sequence(self) -> Optional[str]:
        return self._trav.cdr2_sequence

    @property
    def cdr1b_sequence(self) -> Optional[str]:
        return self._trbv.cdr1_sequence

    @property
    def cdr2b_sequence(self) -> Optional[str]:
        return self._trbv.cdr2_sequence

    @property
    def both_chains_specified(self) -> bool:
        tra_specified = (not self._trav._gene_is_unknown()) or (
            self.junction_a_sequence is not None
        )
        trb_specified = (not self._trbv._gene_is_unknown()) or (
            self.junction_b_sequence is not None
        )
        return tra_specified and trb_specified

    def copy(self) -> "Tcr":
        return copy.deepcopy(self)

    def drop_tra(self) -> None:
        self._trav = Tcrv(None, None)
        self.junction_a_sequence = None

    def drop_trb(self) -> None:
        self._trbv = Tcrv(None, None)
        self.junction_b_sequence = None

    def __eq__(self, __value: object) -> bool:
        return (
            self._trav == __value._trav
            and self.junction_a_sequence == __value.junction_a_sequence
            and self._trbv == __value._trbv
            and self.junction_b_sequence == __value.junction_b_sequence
        )

    def __repr__(self) -> str:
        junction_a_repr = self._represent_junction_sequence(self.junction_a_sequence)
        junction_b_repr = self._represent_junction_sequence(self.junction_b_sequence)

        return (
            f"Tra({self._trav}/{junction_a_repr})/Trb({self._trbv}/{junction_b_repr})"
        )

    def _represent_junction_sequence(self, sequence: Optional[str]) -> str:
        if sequence is None:
            return "?"
        else:
            return sequence


def make_tcr_from_components(
    trav_symbol: Optional[str],
    junction_a_sequence: Optional[str],
    trbv_symbol: Optional[str],
    junction_b_sequence: Optional[str],
) -> Tcr:
    trav = _get_trav_from_symbol(trav_symbol)
    trbv = _get_trbv_from_symbol(trbv_symbol)
    _ensure_valid_junction(junction_a_sequence, "A")
    _ensure_valid_junction(junction_b_sequence, "B")

    return Tcr(
        trav=trav,
        junction_a_sequence=junction_a_sequence,
        trbv=trbv,
        junction_b_sequence=junction_b_sequence,
    )


def _get_trav_from_symbol(symbol: Optional[str]) -> Tcrv:
    if symbol is None:
        return Tcrv(gene=None, allele_num=None)

    gene = _get_trav_gene_object_from_symbol(symbol)
    allele_num = _get_allele_number_from_symbol(symbol)
    return Tcrv(gene=gene, allele_num=allele_num)


def _get_trbv_from_symbol(symbol: Optional[str]) -> Tcrv:
    if symbol is None:
        return Tcrv(gene=None, allele_num=None)

    gene = _get_trbv_gene_object_from_symbol(symbol)
    allele_num = _get_allele_number_from_symbol(symbol)
    return Tcrv(gene=gene, allele_num=allele_num)


def _get_trav_gene_object_from_symbol(symbol: str) -> TravGene:
    str_representing_gene = symbol.split("*")[0]
    if str_representing_gene not in TravGene.__members__:
        raise exception.BadV("A")
    return TravGene[str_representing_gene]


def _get_trbv_gene_object_from_symbol(symbol: str) -> TrbvGene:
    str_representing_gene = symbol.split("*")[0]
    if str_representing_gene not in TrbvGene.__members__:
        raise exception.BadV("B")
    return TrbvGene[str_representing_gene]


def _get_allele_number_from_symbol(symbol: str) -> int:
    split_at_asterisk = symbol.split("*")
    has_allele_number = len(split_at_asterisk) == 2

    if not has_allele_number:
        return None

    str_representing_allele_number = split_at_asterisk[1]

    return int(str_representing_allele_number)


def _ensure_valid_junction(junction: Optional[str], chain: Literal["A", "B"]):
    if junction is None:
        return

    standardised = tt.junction.standardize(junction, strict=False, log_failures=False)

    if standardised is None:
        raise exception.BadJunction(chain)

    if standardised != junction:
        logger.warning(
            f"{junction} doesn't look like a standardised junction. Are you sure you've standardised?"
        )



from typing import Optional


class MhcGene:
    def __init__(self, symbol: Optional[str]) -> None:
        self.symbol = symbol

    def could_be_equal_to(self, other: "MhcGene") -> bool:
        if self.symbol is None or other.symbol is None:
            return True

        self_is_subset_of_other = other.symbol in self.symbol
        other_is_subset_of_self = self.symbol in other.symbol
        return self_is_subset_of_other or other_is_subset_of_self

    def __repr__(self) -> str:
        if self.symbol is None:
            return "?"

        return self.symbol


class Pmhc:
    def __init__(
        self, epitope_sequence: Optional[str], mhc_a: MhcGene, mhc_b: MhcGene
    ) -> None:
        self.epitope_sequence = epitope_sequence
        self.mhc_a = mhc_a
        self.mhc_b = mhc_b

    def __eq__(self, __value: object) -> bool:
        if self.epitope_sequence is None:
            return False

        same_epitope = self.epitope_sequence == __value.epitope_sequence
        maybe_same_mhc_a = self.mhc_a.could_be_equal_to(__value.mhc_a)
        maybe_same_mhc_b = self.mhc_b.could_be_equal_to(__value.mhc_b)
        return same_epitope and maybe_same_mhc_a and maybe_same_mhc_b

    def __repr__(self) -> str:
        epitope_representation = (
            "?" if self.epitope_sequence is None else self.epitope_sequence
        )
        return f"{epitope_representation}/{self.mhc_a}/{self.mhc_b}"

    def __hash__(self) -> int:
        return hash((self.epitope_sequence, self.mhc_a.symbol, self.mhc_b.symbol))


def make_pmhc_from_components(
    epitope_sequence: Optional[str],
    mhc_a_symbol: Optional[str],
    mhc_b_symbol: Optional[str],
) -> Pmhc:
    mhc_a = MhcGene(mhc_a_symbol)
    mhc_b = MhcGene(mhc_b_symbol)
    return Pmhc(epitope_sequence=epitope_sequence, mhc_a=mhc_a, mhc_b=mhc_b)



from abc import ABC, abstractmethod
from libtcrlm.self_attention_stack import SelfAttentionStack
from torch import Tensor
from torch.nn import functional as F


LOCATION_OF_CLS_TOKEN = 0


class VectorRepresentationDelegate(ABC):
    @abstractmethod
    def __init__(self, self_attention_stack: SelfAttentionStack) -> None:
        pass

    @abstractmethod
    def get_vector_representations_of(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        pass


class AveragePoolVectorRepresentationDelegate(VectorRepresentationDelegate):
    def __init__(self, self_attention_stack: SelfAttentionStack) -> None:
        self._self_attention_stack = self_attention_stack

    def get_vector_representations_of(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        embeddings_from_penultimate_layer = (
            self._self_attention_stack.get_token_embeddings_at_penultimate_layer(
                token_embeddings, padding_mask
            )
        )
        padding_and_cls_mask = self._edit_mask_to_also_mark_cls_tokens(padding_mask)

        aa_embeddings_averaged = self._masked_average_pool(
            embeddings_from_penultimate_layer, padding_and_cls_mask
        )

        l2_normed_embeddings = F.normalize(aa_embeddings_averaged, p=2, dim=1)

        return l2_normed_embeddings

    def _edit_mask_to_also_mark_cls_tokens(self, padding_mask: Tensor) -> Tensor:
        padding_mask_with_cls_also_masked = padding_mask.clone()
        padding_mask_with_cls_also_masked[:, 0] = 1

        return padding_mask_with_cls_also_masked

    def _masked_average_pool(
        self, token_embeddings: Tensor, padding_and_cls_mask: Tensor
    ) -> Tensor:
        zero_where_cls_and_padding = padding_and_cls_mask.logical_not()
        mask_broadcastable_with_embeddings = zero_where_cls_and_padding.unsqueeze(-1)

        only_non_cls_embeddings = token_embeddings * mask_broadcastable_with_embeddings
        token_embeddings_summed = only_non_cls_embeddings.sum(1)
        token_embeddings_averaged = (
            token_embeddings_summed / mask_broadcastable_with_embeddings.sum(1)
        )

        return token_embeddings_averaged


class ClsVectorRepresentationDelegate(VectorRepresentationDelegate):
    def __init__(self, self_attention_stack: SelfAttentionStack) -> None:
        self._self_attention_stack = self_attention_stack

    def get_vector_representations_of(
        self, token_embeddings: Tensor, padding_mask: Tensor
    ) -> Tensor:
        final_token_embeddings = self._self_attention_stack.forward(
            token_embeddings, padding_mask
        )
        final_cls_embeddings = final_token_embeddings[:, LOCATION_OF_CLS_TOKEN, :]
        l2_normed_cls_embeddings = F.normalize(final_cls_embeddings, p=2, dim=1)

        return l2_normed_cls_embeddings



from abc import ABC, abstractmethod
from libtcrlm.schema import AminoAcid
from torch import Tensor
from torch.nn import Linear, Module


class MlmTokenPredictionProjector(ABC, Module):
    @abstractmethod
    def forward(self, token_embeddings: Tensor) -> Tensor:
        pass


class AminoAcidTokenProjector(MlmTokenPredictionProjector):
    def __init__(self, d_model: int) -> None:
        super().__init__()

        self.projector = Linear(in_features=d_model, out_features=len(AminoAcid))

    def forward(self, token_embeddings: Tensor) -> Tensor:
        return self.projector.forward(token_embeddings)



from enum import Enum


class AminoAcid(Enum):
    A = 0
    C = 1
    D = 2
    E = 3
    F = 4
    G = 5
    H = 6
    I = 7
    K = 8
    L = 9
    M = 10
    N = 11
    P = 12
    Q = 13
    R = 14
    S = 15
    T = 16
    V = 17
    W = 18
    Y = 19



import math
from libtcrlm.tokeniser.token_indices import (
    DefaultTokenIndex,
    AminoAcidTokenIndex,
    CdrCompartmentIndex,
    SingleChainCdrCompartmentIndex,
)
from libtcrlm.token_embedder.token_embedder import TokenEmbedder
from libtcrlm.token_embedder.simple_relative_position_embedding import (
    SimpleRelativePositionEmbedding,
)
from libtcrlm.token_embedder.sin_position_embedding import SinPositionEmbedding
from libtcrlm.token_embedder.one_hot_token_index_embedding import (
    OneHotTokenIndexEmbedding,
)
from libtcrlm.token_embedder.blosum_embedding import BlosumEmbedding
import torch
from torch import FloatTensor, LongTensor
from torch.nn import Embedding


MAX_PLAUSIBLE_CDR_LENGTH = 100


class CdrBlosumEmbedder(TokenEmbedder):
    def __init__(self) -> None:
        super().__init__()
        self._blosum_embedding = BlosumEmbedding()
        self._position_embedding = SimpleRelativePositionEmbedding()
        self._compartment_embedding = OneHotTokenIndexEmbedding(CdrCompartmentIndex)

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self._blosum_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self._position_embedding.forward(tokenised_tcrs[:, :, 1:3])
        compartment_component = self._compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )
        all_components_stacked = torch.concatenate(
            [token_component, position_component, compartment_component], dim=-1
        )
        return all_components_stacked


class CdrSimpleEmbedder(TokenEmbedder):
    def __init__(self) -> None:
        super().__init__()
        self._token_embedding = OneHotTokenIndexEmbedding(AminoAcidTokenIndex)
        self._position_embedding = SimpleRelativePositionEmbedding()
        self._compartment_embedding = OneHotTokenIndexEmbedding(CdrCompartmentIndex)

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self._token_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self._position_embedding.forward(tokenised_tcrs[:, :, 1:3])
        compartment_component = self._compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )
        all_components_stacked = torch.concatenate(
            [token_component, position_component, compartment_component], dim=-1
        )
        return all_components_stacked


class CdrEmbedder(TokenEmbedder):
    def __init__(self, embedding_dim: int) -> None:
        super().__init__()

        self._embedding_dim = embedding_dim

        self.token_embedding = Embedding(
            num_embeddings=len(AminoAcidTokenIndex),
            embedding_dim=embedding_dim,
            padding_idx=DefaultTokenIndex.NULL,
        )
        self.position_embedding = SinPositionEmbedding(
            num_embeddings=MAX_PLAUSIBLE_CDR_LENGTH, embedding_dim=embedding_dim
        )
        self.compartment_embedding = Embedding(
            num_embeddings=len(CdrCompartmentIndex),
            embedding_dim=embedding_dim,
            padding_idx=DefaultTokenIndex.NULL,
        )

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self.token_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self.position_embedding.forward(tokenised_tcrs[:, :, 1])
        compartment_component = self.compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )

        all_components_summed = (
            token_component + position_component + compartment_component
        )

        return all_components_summed * math.sqrt(self._embedding_dim)


class SingleChainCdrEmbedder(TokenEmbedder):
    def __init__(self, embedding_dim: int) -> None:
        super().__init__()

        self._embedding_dim = embedding_dim

        self.token_embedding = Embedding(
            num_embeddings=len(AminoAcidTokenIndex),
            embedding_dim=embedding_dim,
            padding_idx=DefaultTokenIndex.NULL,
        )
        self.position_embedding = SinPositionEmbedding(
            num_embeddings=MAX_PLAUSIBLE_CDR_LENGTH, embedding_dim=embedding_dim
        )
        self.compartment_embedding = Embedding(
            num_embeddings=len(SingleChainCdrCompartmentIndex),
            embedding_dim=embedding_dim,
            padding_idx=DefaultTokenIndex.NULL,
        )

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self.token_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self.position_embedding.forward(tokenised_tcrs[:, :, 1])
        compartment_component = self.compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )

        all_components_summed = (
            token_component + position_component + compartment_component
        )

        return all_components_summed * math.sqrt(self._embedding_dim)


class SingleChainCdrEmbedderWithRelativePositions(TokenEmbedder):
    def __init__(self, embedding_dim: int) -> None:
        super().__init__()

        self._embedding_dim = embedding_dim

        self._token_embedding = Embedding(
            num_embeddings=len(AminoAcidTokenIndex),
            embedding_dim=embedding_dim,
            padding_idx=DefaultTokenIndex.NULL,
        )
        self._position_embedding = SimpleRelativePositionEmbedding()
        self._compartment_embedding = OneHotTokenIndexEmbedding(
            SingleChainCdrCompartmentIndex
        )

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self._token_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self._position_embedding.forward(tokenised_tcrs[:, :, 1:3])
        compartment_component = self._compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )
        all_components_stacked = torch.concatenate(
            [token_component, position_component, compartment_component], dim=-1
        )
        return all_components_stacked * math.sqrt(self._embedding_dim)


class SingleChainCdrSimpleEmbedder(TokenEmbedder):
    def __init__(self) -> None:
        super().__init__()
        self._token_embedding = OneHotTokenIndexEmbedding(AminoAcidTokenIndex)
        self._position_embedding = SimpleRelativePositionEmbedding()
        self._compartment_embedding = OneHotTokenIndexEmbedding(
            SingleChainCdrCompartmentIndex
        )

    def forward(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        token_component = self._token_embedding.forward(tokenised_tcrs[:, :, 0])
        position_component = self._position_embedding.forward(tokenised_tcrs[:, :, 1:3])
        compartment_component = self._compartment_embedding.forward(
            tokenised_tcrs[:, :, 3]
        )
        all_components_stacked = torch.concatenate(
            [token_component, position_component, compartment_component], dim=-1
        )
        return all_components_stacked



from enum import IntEnum
import torch
from torch import FloatTensor, LongTensor
from torch.nn import Module


class OneHotTokenIndexEmbedding(Module):
    def __init__(self, token_index: IntEnum) -> None:
        super().__init__()
        self._register_token_embeddings(token_index)

    def _register_token_embeddings(self, token_index: IntEnum) -> FloatTensor:
        num_tokens = len(token_index)
        num_tokens_excluding_null = num_tokens - 1

        null_embedding = torch.zeros((1, num_tokens_excluding_null))
        non_null_token_embeddings = torch.eye(num_tokens_excluding_null)
        token_embeddings = torch.concatenate(
            [null_embedding, non_null_token_embeddings], dim=0
        )

        self.register_buffer("_token_embeddings", token_embeddings)

    def forward(self, token_indices: LongTensor) -> FloatTensor:
        return self._token_embeddings[token_indices]



from .token_embedder import TokenEmbedder
from .cdr3_embedder import (
    Cdr3Embedder,
    Cdr3EmbedderWithRelativePositions,
    Cdr3SimpleEmbedder,
    SingleChainCdr3SimpleEmbedder,
)
from .cdr_embedder import (
    CdrBlosumEmbedder,
    CdrSimpleEmbedder,
    CdrEmbedder,
    SingleChainCdrEmbedder,
    SingleChainCdrEmbedderWithRelativePositions,
    SingleChainCdrSimpleEmbedder,
)



from libtcrlm.tokeniser.token_indices import DefaultTokenIndex
from libtcrlm.token_embedder.token_embedder import TokenEmbedder
from libtcrlm.mlm_token_prediction_projector import MlmTokenPredictionProjector
from libtcrlm.self_attention_stack import SelfAttentionStack
from libtcrlm.vector_representation_delegate import VectorRepresentationDelegate
from torch import BoolTensor, FloatTensor, LongTensor
from torch.nn import Module


class Bert(Module):
    def __init__(
        self,
        token_embedder: TokenEmbedder,
        self_attention_stack: SelfAttentionStack,
        mlm_token_prediction_projector: MlmTokenPredictionProjector,
        vector_representation_delegate: VectorRepresentationDelegate,
    ) -> None:
        super().__init__()

        self._token_embedder = token_embedder
        self._self_attention_stack = self_attention_stack
        self._mlm_token_prediction_projector = mlm_token_prediction_projector
        self._vector_representation_delegate = vector_representation_delegate

    @property
    def d_model(self) -> int:
        return self._self_attention_stack.d_model

    def set_fine_tuning_mode(self, turn_on: bool) -> None:
        self._self_attention_stack.set_fine_tuning_mode(turn_on)

    def get_vector_representations_of(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        raw_token_embeddings = self._embed(tokenised_tcrs)
        padding_mask = self._get_padding_mask(tokenised_tcrs)
        vector_representations = (
            self._vector_representation_delegate.get_vector_representations_of(
                raw_token_embeddings, padding_mask
            )
        )

        return vector_representations

    def get_mlm_token_predictions_for(
        self, tokenised_and_masked_tcrs: LongTensor
    ) -> FloatTensor:
        raw_token_embeddings = self._embed(tokenised_and_masked_tcrs)
        padding_mask = self._get_padding_mask(tokenised_and_masked_tcrs)
        contextualised_token_embeddings = self._self_attention_stack.forward(
            raw_token_embeddings, padding_mask
        )
        mlm_token_predictions = self._mlm_token_prediction_projector.forward(
            contextualised_token_embeddings
        )

        return mlm_token_predictions

    def _embed(self, tokenised_tcrs: LongTensor) -> FloatTensor:
        return self._token_embedder.forward(tokenised_tcrs)

    def _get_padding_mask(self, tokenised_tcrs: LongTensor) -> BoolTensor:
        return tokenised_tcrs[:, :, 0] == DefaultTokenIndex.NULL



from typing import Literal


class BadV(Exception):
    def __init__(self, chain: Literal["A", "B"]):
        self.chain = chain


class BadJunction(Exception):
    def __init__(self, chain: Literal["A", "B"]):
        self.chain = chain



import blosum
import itertools
from libtcrlm.tokeniser.token_indices import DefaultTokenIndex
from libtcrlm.schema import AminoAcid
import torch
from torch import FloatTensor, LongTensor
from torch.nn import Module, Embedding


class BlosumEmbedding(Module):
    def __init__(self) -> None:
        super().__init__()

        self._special_token_embeddings = Embedding(
            num_embeddings=len(DefaultTokenIndex),
            embedding_dim=len(AminoAcid),
            padding_idx=DefaultTokenIndex.NULL,
        )
        self._register_blosum_embeddings()

    def _register_blosum_embeddings(self) -> None:
        blosum_matrix = blosum.BLOSUM(62)

        null_embedding = torch.zeros(size=(len(DefaultTokenIndex), len(AminoAcid)))
        aa_embeddings = torch.zeros(size=(len(AminoAcid), len(AminoAcid)))

        for row, column in itertools.product(AminoAcid, repeat=2):
            blosum_score = blosum_matrix[row.name][column.name]
            aa_embeddings[row.value, column.value] = blosum_score

        blosum_embeddings = torch.concatenate([null_embedding, aa_embeddings], dim=0)
        blosum_embeddings_normalised = blosum_embeddings / blosum_embeddings.abs().max()

        self.register_buffer("_blosum_embeddings", blosum_embeddings_normalised)

    def forward(self, token_indices: LongTensor) -> FloatTensor:
        special_token_mask = token_indices < len(DefaultTokenIndex)
        token_indices_aa_masked_out = token_indices * special_token_mask

        special_token_embeddings = self._special_token_embeddings.forward(
            token_indices_aa_masked_out
        )
        aa_blosum_embeddings = self._blosum_embeddings[token_indices]

        return special_token_embeddings + aa_blosum_embeddings



from .amino_acid import AminoAcid
from .tcr import Tcr, make_tcr_from_components
from .pmhc import Pmhc, make_pmhc_from_components
from .tcr_pmhc_pair import TcrPmhcPair
from .df_manipulation import generate_tcr_pmhc_series, generate_tcr_series



from libtcrlm.schema import Tcr
from libtcrlm.schema import Pmhc


class TcrPmhcPair:
    def __init__(self, tcr: Tcr, pmhc: Pmhc) -> None:
        self.tcr = tcr
        self.pmhc = pmhc

    def __repr__(self) -> str:
        return f"({self.tcr}) - ({self.pmhc})"



from libtcrlm import schema
from libtcrlm.schema import exception
from libtcrlm.schema import Tcr, TcrPmhcPair
import pandas as pd
from pandas import DataFrame, Series


def generate_tcr_pmhc_series(data: DataFrame) -> Series:
    tcr_pmhc_series = data.apply(_generate_tcr_pmhc_pair_from_row, axis="columns")
    return tcr_pmhc_series


def generate_tcr_series(data: DataFrame) -> Series:
    tcr_series = data.apply(_generate_tcr_from_row, axis="columns")
    return tcr_series


def _generate_tcr_pmhc_pair_from_row(row: Series) -> TcrPmhcPair:
    tcr = _generate_tcr_from_row(row)

    epitope = _get_value_if_not_na_else_none(row.Epitope)
    mhc_a = _get_value_if_not_na_else_none(row.MHCA)
    mhc_b = _get_value_if_not_na_else_none(row.MHCB)

    try:
        pmhc = schema.make_pmhc_from_components(
            epitope_sequence=epitope, mhc_a_symbol=mhc_a, mhc_b_symbol=mhc_b
        )
    except Exception:
        raise ValueError(
            f"Bad pMHC data at index {row.name}. "
            "Most likely, there is a non-standard MH symbol, or an invalid peptide sequence. "
            "You can use tidytcells with specific flags to filter out non-valid data (see https://tidytcells.readthedocs.io).\n\n"
            f"Epitope: {row.Epitope}\n"
            f"MHCA:    {row.MHCA}\n"
            f"MHCB:    {row.MHCB}"
        )

    return TcrPmhcPair(tcr, pmhc)


def _generate_tcr_from_row(row: Series) -> Tcr:
    trav = _get_value_if_not_na_else_none(row.TRAV)
    trbv = _get_value_if_not_na_else_none(row.TRBV)
    junction_a = _get_value_if_not_na_else_none(row.CDR3A)
    junction_b = _get_value_if_not_na_else_none(row.CDR3B)

    try:
        return schema.make_tcr_from_components(
            trav_symbol=trav,
            junction_a_sequence=junction_a,
            trbv_symbol=trbv,
            junction_b_sequence=junction_b,
        )
    except exception.BadV as e:
        colname = f"TR{e.chain}V"
        bad_symbol = row[colname]
        raise ValueError(
            f"Bad {colname} symbol at index {row.name}: {bad_symbol}. "
            "Have you ensured that all TR V symbols are standardised and functional? "
            "You can use tidytcells with specific flags to filter out non-valid data (see https://tidytcells.readthedocs.io)."
        )
    except exception.BadJunction as e:
        colname = f"CDR3{e.chain}"
        bad_junction = row[colname]
        raise ValueError(
            f"Bad {colname} sequence at index {row.name}: {bad_junction}. "
            "Have you ensured that all CDR3 sequences are standardised and valid? "
            "You can use tidytcells with specific flags to filter out non-valid data (see https://tidytcells.readthedocs.io)."
        )


def _get_value_if_not_na_else_none(value) -> any:
    if pd.isna(value):
        return None
    return value



import pytest
import torch
from libtcrlm import schema
from libtcrlm.tokeniser import *


class TestBetaCdr3Tokeniser:
    tokeniser = BetaCdr3Tokeniser()

    def test_tokenise(self, mock_tcr):
        tokenised_tcr = self.tokeniser.tokenise(mock_tcr)
        expected = torch.tensor(
            [
                [2, 0, 0],
                [4, 1, 6],
                [3, 2, 6],
                [18, 3, 6],
                [16, 4, 6],
                [22, 5, 6],
                [7, 6, 6],
            ]
        )

        assert torch.equal(tokenised_tcr, expected)

    def test_tokenise_tcr_with_empty_beta_junction(self):
        tcr_with_empty_beta = schema.make_tcr_from_components(
            "TRAV1-1*01", "CASQYF", "TRBV2*01", None
        )

        with pytest.raises(RuntimeError):
            self.tokeniser.tokenise(tcr_with_empty_beta)


def test_cdr_tokeniser(mock_tcr):
    tokenised_tcr = CdrTokeniser().tokenise(mock_tcr)
    expected = torch.tensor(
        [
            [2, 0, 0, 0],
            [19, 1, 6, 1],
            [18, 2, 6, 1],
            [8, 3, 6, 1],
            [7, 4, 6, 1],
            [22, 5, 6, 1],
            [8, 6, 6, 1],
            [14, 1, 6, 2],
            [3, 2, 6, 2],
            [12, 3, 6, 2],
            [5, 4, 6, 2],
            [8, 5, 6, 2],
            [12, 6, 6, 2],
            [4, 1, 6, 3],
            [3, 2, 6, 3],
            [19, 3, 6, 3],
            [16, 4, 6, 3],
            [22, 5, 6, 3],
            [7, 6, 6, 3],
            [18, 1, 5, 4],
            [14, 2, 5, 4],
            [9, 3, 5, 4],
            [12, 4, 5, 4],
            [22, 5, 5, 4],
            [7, 1, 6, 5],
            [22, 2, 6, 5],
            [14, 3, 6, 5],
            [14, 4, 6, 5],
            [6, 5, 6, 5],
            [10, 6, 6, 5],
            [4, 1, 6, 6],
            [3, 2, 6, 6],
            [18, 3, 6, 6],
            [16, 4, 6, 6],
            [22, 5, 6, 6],
            [7, 6, 6, 6],
        ]
    )

    assert torch.equal(tokenised_tcr, expected)


class TestBetaCdrTokeniser:
    tokeniser = BetaCdrTokeniser()

    def test_tokenise(self, mock_tcr):
        tokenised_tcr = self.tokeniser.tokenise(mock_tcr)
        expected = torch.tensor(
            [
                [2, 0, 0, 0],
                [18, 1, 5, 1],
                [14, 2, 5, 1],
                [9, 3, 5, 1],
                [12, 4, 5, 1],
                [22, 5, 5, 1],
                [7, 1, 6, 2],
                [22, 2, 6, 2],
                [14, 3, 6, 2],
                [14, 4, 6, 2],
                [6, 5, 6, 2],
                [10, 6, 6, 2],
                [4, 1, 6, 3],
                [3, 2, 6, 3],
                [18, 3, 6, 3],
                [16, 4, 6, 3],
                [22, 5, 6, 3],
                [7, 6, 6, 3],
            ]
        )

        assert torch.equal(tokenised_tcr, expected)

    def test_tokenise_tcr_with_empty_beta_junction(self):
        tcr_with_empty_beta = schema.make_tcr_from_components(
            "TRAV1-1*01", "CASQYF", None, None
        )

        with pytest.raises(RuntimeError):
            self.tokeniser.tokenise(tcr_with_empty_beta)



from types import ModuleType

import libtcrlm.tokeniser as tokeniser_module
import libtcrlm.self_attention_stack as self_attention_stack_module
import libtcrlm.mlm_token_prediction_projector as mlm_token_prediction_projector_module
import libtcrlm.vector_representation_delegate as vector_representation_delegate_module
import libtcrlm.token_embedder as token_embedder_module

from libtcrlm.tokeniser import Tokeniser
from libtcrlm.bert import Bert
from libtcrlm.token_embedder.token_embedder import TokenEmbedder
from libtcrlm.self_attention_stack import SelfAttentionStack
from libtcrlm.mlm_token_prediction_projector import MlmTokenPredictionProjector
from libtcrlm.vector_representation_delegate import VectorRepresentationDelegate


class ConfigReader:
    def __init__(self, config: dict) -> None:
        self._config = config

    def get_model_name(self) -> str:
        return self._config["model"]["name"]

    def get_config(self) -> dict:
        return self._config

    def get_tokeniser(self) -> Tokeniser:
        config = self._config["data"]["tokeniser"]
        return self._get_object_from_module_using_config(tokeniser_module, config)

    def get_bert(self) -> Bert:
        token_embedder = self._get_token_embedder()
        self_attention_stack = self._get_self_attention_stack()
        mlm_token_prediction_projector = self._get_mlm_token_prediction_projector()
        vector_representation_delegate = (
            self._get_vector_representation_delegate_for_self_attention_stack(
                self_attention_stack
            )
        )

        bert = Bert(
            token_embedder=token_embedder,
            self_attention_stack=self_attention_stack,
            mlm_token_prediction_projector=mlm_token_prediction_projector,
            vector_representation_delegate=vector_representation_delegate,
        )

        return bert

    def _get_token_embedder(self) -> TokenEmbedder:
        config = self._config["model"]["token_embedder"]
        return self._get_object_from_module_using_config(token_embedder_module, config)

    def _get_self_attention_stack(self) -> SelfAttentionStack:
        config = self._config["model"]["self_attention_stack"]
        return self._get_object_from_module_using_config(
            self_attention_stack_module, config
        )

    def _get_mlm_token_prediction_projector(
        self,
    ) -> MlmTokenPredictionProjector:
        config = self._config["model"]["mlm_token_prediction_projector"]
        return self._get_object_from_module_using_config(
            mlm_token_prediction_projector_module, config
        )

    def _get_vector_representation_delegate_for_self_attention_stack(
        self, self_attention_stack: SelfAttentionStack
    ) -> VectorRepresentationDelegate:
        config = self._config["model"]["vector_representation_delegate"]
        class_name = config["class"]
        initargs = config["initargs"]
        VectorRepresentationDelegateClass = getattr(
            vector_representation_delegate_module, class_name
        )
        return VectorRepresentationDelegateClass(
            self_attention_stack=self_attention_stack, **initargs
        )

    def _get_object_from_module_using_config(
        self, module: ModuleType, config: dict
    ) -> any:
        class_name = config["class"]
        initargs = config["initargs"]
        Class = getattr(module, class_name)
        return Class(**initargs)



from abc import ABC, abstractmethod
from torch import Tensor
from torch.nn import Module


class TokenEmbedder(ABC, Module):
    @abstractmethod
    def forward(self, tokenised_tcrs: Tensor) -> Tensor:
        pass



from enum import IntEnum


class DefaultTokenIndex(IntEnum):
    NULL = 0
    MASK = 1
    CLS = 2


class AminoAcidTokenIndex(IntEnum):
    NULL = DefaultTokenIndex.NULL
    MASK = DefaultTokenIndex.MASK
    CLS = DefaultTokenIndex.CLS
    A = 3
    C = 4
    D = 5
    E = 6
    F = 7
    G = 8
    H = 9
    I = 10
    K = 11
    L = 12
    M = 13
    N = 14
    P = 15
    Q = 16
    R = 17
    S = 18
    T = 19
    V = 20
    W = 21
    Y = 22


class Cdr3CompartmentIndex(IntEnum):
    NULL = DefaultTokenIndex.NULL
    CDR3A = 1
    CDR3B = 2


class SingleChainCdrCompartmentIndex(IntEnum):
    NULL = DefaultTokenIndex.NULL
    CDR1 = 1
    CDR2 = 2
    CDR3 = 3


class CdrCompartmentIndex(IntEnum):
    NULL = DefaultTokenIndex.NULL
    CDR1A = 1
    CDR2A = 2
    CDR3A = 3
    CDR1B = 4
    CDR2B = 5
    CDR3B = 6



from libtcrlm.tokeniser.tokeniser import Tokeniser
from libtcrlm.tokeniser.token_indices import AminoAcidTokenIndex, Cdr3CompartmentIndex
from libtcrlm.schema import Tcr
import torch
from torch import Tensor
from typing import List, Optional, Tuple


class Cdr3Tokeniser(Tokeniser):
    """
    Tokenise TCR in terms of its alpha and beta chain CDR3s.

    Dim 0: token index
    Dim 1: token position
    Dim 2: CDR length
    Dim 3: CDR index
    """

    token_vocabulary_index = AminoAcidTokenIndex

    def tokenise(self, tcr: Tcr) -> Tensor:
        initial_cls_vector = (AminoAcidTokenIndex.CLS, 0, 0, Cdr3CompartmentIndex.NULL)

        cdr3a = self._convert_to_numerical_form(
            tcr.junction_a_sequence, Cdr3CompartmentIndex.CDR3A
        )
        cdr3b = self._convert_to_numerical_form(
            tcr.junction_b_sequence, Cdr3CompartmentIndex.CDR3B
        )

        all_cdrs_tokenised = [initial_cls_vector] + cdr3a + cdr3b

        number_of_tokens_other_than_initial_cls = len(all_cdrs_tokenised) - 1
        if number_of_tokens_other_than_initial_cls == 0:
            raise RuntimeError(f"tcr {tcr} does not contain any TCR information")

        return torch.tensor(all_cdrs_tokenised, dtype=torch.long)

    def _convert_to_numerical_form(
        self, aa_sequence: Optional[str], cdr_index: Cdr3CompartmentIndex
    ) -> List[Tuple[int]]:
        if aa_sequence is None:
            return []

        token_indices = [AminoAcidTokenIndex[aa] for aa in aa_sequence]
        token_positions = [idx for idx, _ in enumerate(aa_sequence, start=1)]
        cdr_length = [len(aa_sequence) for _ in aa_sequence]
        compartment_index = [cdr_index for _ in aa_sequence]

        iterator_over_token_vectors = zip(
            token_indices, token_positions, cdr_length, compartment_index
        )

        return list(iterator_over_token_vectors)


class BetaCdr3Tokeniser(Tokeniser):
    """
    Tokenise TCR in terms of its beta chain CDR3.

    Dim 0: token index
    Dim 1: token position
    Dim 2: CDR length
    """

    token_vocabulary_index = AminoAcidTokenIndex

    def tokenise(self, tcr: Tcr) -> Tensor:
        initial_cls_vector = (
            AminoAcidTokenIndex.CLS,
            0,
            0,
        )
        cdr3b = self._convert_to_numerical_form(tcr.junction_b_sequence)

        tokenised = [initial_cls_vector] + cdr3b

        number_of_tokens_other_than_initial_cls = len(tokenised) - 1
        if number_of_tokens_other_than_initial_cls == 0:
            raise RuntimeError(f"tcr {tcr} does not contain beta junction information")

        return torch.tensor(tokenised, dtype=torch.long)

    def _convert_to_numerical_form(
        self, aa_sequence: Optional[str]
    ) -> List[Tuple[int]]:
        if aa_sequence is None:
            return []

        token_indices = [AminoAcidTokenIndex[aa] for aa in aa_sequence]
        token_positions = [idx for idx, _ in enumerate(aa_sequence, start=1)]
        cdr_length = [len(aa_sequence) for _ in aa_sequence]

        iterator_over_token_vectors = zip(token_indices, token_positions, cdr_length)

        return list(iterator_over_token_vectors)



import pytest
import torch
from torch import Tensor

from libtcrlm.bert import Bert


BATCH_SIZE = 5
TOKENISED_TCR_LENGTH = 5
TOKEN_DIMENSIONALITY = 4

AMINO_ACID_VOCABULARY_SIZE = 20


@pytest.fixture
def mock_tokenised_tcrs():
    return torch.ones(
        (BATCH_SIZE, TOKENISED_TCR_LENGTH, TOKEN_DIMENSIONALITY), dtype=torch.long
    )


def test_get_vector_representations_of(
    toy_bert: Bert, toy_bert_d_model, mock_tokenised_tcrs
):
    result = toy_bert.get_vector_representations_of(mock_tokenised_tcrs)

    assert type(result) == Tensor
    assert result.dim() == 2
    assert result.size(0) == BATCH_SIZE
    assert result.size(1) == toy_bert_d_model


def test_get_mlm_token_predictions_for(toy_bert: Bert, mock_tokenised_tcrs):
    result = toy_bert.get_mlm_token_predictions_for(mock_tokenised_tcrs)

    assert type(result) == Tensor
    assert result.dim() == 3
    assert result.size(0) == BATCH_SIZE
    assert result.size(1) == TOKENISED_TCR_LENGTH
    assert result.size(2) == AMINO_ACID_VOCABULARY_SIZE



from torch import FloatTensor, LongTensor
from torch.nn import Module

from libtcrlm.tokeniser.token_indices import DefaultTokenIndex


class SimpleRelativePositionEmbedding(Module):
    def forward(self, position_indices: LongTensor) -> FloatTensor:
        """
        Input tensor should have shape (..., 2) with first of the final two dimensions encoding token position, and the final dimension encoding compartment length.
        """
        null_mask = position_indices[..., 0] == DefaultTokenIndex.NULL
        zero_indexed_token_positions = position_indices[..., 0] - 1
        compartment_length_minus_one = position_indices[..., 1] - 1

        relative_token_positions = (
            zero_indexed_token_positions / compartment_length_minus_one
        )

        RELATIVE_POSITION_IF_ONLY_ONE_TOKEN_IN_COMPARTMENT = 0.5
        relative_token_positions[relative_token_positions.isnan()] = (
            RELATIVE_POSITION_IF_ONLY_ONE_TOKEN_IN_COMPARTMENT
        )
        relative_token_positions[null_mask] = 0

        relative_token_positions = relative_token_positions.unsqueeze(dim=-1)

        return relative_token_positions
